//============================================================================
// Bare-Bones Tablet Interface for C++
//
//  Wintab Implementation file
//
//  Author:        William Baxter (baxter@cs.unc.edu)
//  Last Modified: Dec 2005
//============================================================================
// (MIT License)
// Copyright (c) 2002--2009 William Baxter
// 
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
// BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
// ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//============================================================================

# define WIN32_LEAN_AND_MEAN
# include <windows.h>

// Set to 1 to get more debug output
#define VERBOSE_DEBUG 0

/* Because we'll look up functions from the DLL, don't prototype them */
#define NOWTFUNCTIONS
#include <wintab.h>

// We want it all!
#define PACKETNAME FULL
#define FULLPACKETDATA	\
  (   PK_X | PK_Y | PK_Z | PK_BUTTONS \
  | PK_CONTEXT \
  | PK_STATUS  \
  | PK_TIME    \
  | PK_CHANGED \
  | PK_CURSOR \
  | PK_NORMAL_PRESSURE \
  | PK_TANGENT_PRESSURE /* the scroll wheel on wacom airbrush */\
  | PK_ORIENTATION \
  | PK_ROTATION \
  )
#define FULLPACKETMODE	0 /* ALL FIELDS ABSOLUTE */
#define FULLPACKETTILT	PKEXT_ABSOLUTE
#include <pktdef.h>
#undef PACKETNAME

#include <sys/timeb.h> // for waitForNextEvent
#include <vector>
#include <string>
#include <cassert>
using std::vector;
using std::string;

#include <math.h>
#include <stdio.h>

#include "bbtabwin.h"
#include "bbtablet.h"


#if defined(DEBUG) || defined(_DEBUG)
  static void DebugMsg(const char *fmt, ...) 
  {
    va_list args;
    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args); 
  }
#if VERBOSE_DEBUG
  #define VerboseDebugMsg DebugMsg
#else
  inline void VerboseDebugMsg(const char*fmt,...) {}
#endif
#else /* NONDEBUG */
  inline void DebugMsg(const char *fmt, ...) { }
  inline void VerboseDebugMsg(const char*fmt,...) {}
#endif

#ifndef M_PI
static const double M_PI = 3.1415926535897932;
#endif
static const float M_PIf = float(M_PI);

inline float fix2float(FIX32 arg)
{
  return arg * (1.0f/65536.0f);
}


//============================================================================
// WINTAB DLL INITIALIZATION
//============================================================================

static PFNWTINFO pfWTInfo = 0;
static PFNWTOPEN pfWTOpen = 0;
static PFNWTCLOSE pfWTClose = 0;
static PFNWTGET  pfWTGet  = 0;
static PFNWTSET  pfWTSet  = 0;
static PFNWTPACKET pfWTPacket = 0;
static PFNWTPACKETSGET pfWTPacketsGet = 0;
static PFNWTQUEUEPACKETSEX pfWTQueuePacketsEx = 0;
static PFNWTOVERLAP pfWTOverlap = 0;

static int initWintabDLL()
{
  HINSTANCE dllinst = LoadLibrary("wintab32.dll");
  if  (!dllinst) {
    // couldn't find wintab library, return failure code.
    return 1;
  }

#ifdef UNICODE
  pfWTInfo =   (PFNWTINFO) GetProcAddress(dllinst, "WTInfoW");
  pfWTOpen =   (PFNWTOPEN) GetProcAddress(dllinst, "WTOpenW");
  pfWTGet  =    (PFNWTGET) GetProcAddress(dllinst, "WTGetW");
  pfWTSet  =    (PFNWTSET) GetProcAddress(dllinst, "WTSetW");
#else /* !UNICODE */
  pfWTInfo =   (PFNWTINFO) GetProcAddress(dllinst, "WTInfoA");
  pfWTOpen =   (PFNWTOPEN) GetProcAddress(dllinst, "WTOpenA");
  pfWTGet  =    (PFNWTGET) GetProcAddress(dllinst, "WTGetA");
  pfWTSet  =    (PFNWTSET) GetProcAddress(dllinst, "WTSetA");
#endif

  // these don't have unicode / nonunicode versions
  pfWTPacket = 
    (PFNWTPACKET) GetProcAddress(dllinst, "WTPacket");
  pfWTPacketsGet = 
    (PFNWTPACKETSGET) GetProcAddress(dllinst, "WTPacketsGet");
  pfWTQueuePacketsEx = 
    (PFNWTQUEUEPACKETSEX) GetProcAddress(dllinst, "WTQueuePacketsEx");
  pfWTClose =   (PFNWTCLOSE) GetProcAddress(dllinst, "WTClose");
  pfWTOverlap = (PFNWTOVERLAP) GetProcAddress(dllinst, "WTOverlap");

  if ( pfWTInfo  &&
       pfWTOpen  &&
       pfWTClose &&
       pfWTGet   &&
       pfWTSet   &&
       pfWTPacketsGet &&
       pfWTQueuePacketsEx &&
       pfWTOverlap )
  {
    // All functions found, OK, Let's Rock
    return 0;
  }

  return 2; // some functions not located.
}

static BOOL CALLBACK monitorEnumCallback(
                                  HMONITOR hMonitor,  // handle to display monitor
                                  HDC hdc1,           // handle to monitor DC
                                  LPRECT lprcMonitor, // monitor intersection rectangle
                                  LPARAM data         // data
                                  )
{
  RECT rc = *lprcMonitor;
  MONITORINFO info; info.cbSize = sizeof(info);
  GetMonitorInfo(hMonitor, &info);


  return 1;
}
#if 0
HMONITOR MonitorFromPoint(
  POINT pt,      // point in virtual desk coords
  DWORD dwFlags  // determine return value
);

#endif


void _bbTabletGetDisplayInformation()
{
  EnumDisplayMonitors(NULL,NULL, monitorEnumCallback, 0);
  int vsx = GetSystemMetrics(SM_XVIRTUALSCREEN);
  int vsy = GetSystemMetrics(SM_YVIRTUALSCREEN);
  int vsw = GetSystemMetrics(SM_CXVIRTUALSCREEN);
  int vsh = GetSystemMetrics(SM_CYVIRTUALSCREEN);
}


//============================================================================
// BBTABLET DEVICE METHODS
//============================================================================


bool bbTablet::initSystem( void *sys_win, int mode )
{
  if (!pfWTInfo) {
    DebugMsg("You've got no Wintab32.dll\n");
    return false;
  }

  // Find out about our tablet
  if (!pfWTInfo(0,0,NULL)) {
    // no tablet or no tablet support
    DebugMsg("No tablet or no tablet support\n");
    return false;
  }

  m_data->m_hwnd = (HWND)sys_win;

  TCHAR strbuf[1024];

#if VERBOSE_DEBUG
  /////////////////////////////////
  // Tablet info
  int nCursors = 0;
  {
    UINT ret = pfWTInfo(WTI_INTERFACE, IFC_WINTABID, strbuf);
    DebugMsg("Tablet interface provider: %s\n", LPCSTR(strbuf));
    UINT ntabs;
    UINT val;
    ret = pfWTInfo(WTI_INTERFACE, IFC_NDEVICES, &ntabs);
    DebugMsg("# of devices present: %d\n", ntabs);
    if (ret = pfWTInfo(WTI_INTERFACE, IFC_NCURSORS, &nCursors))
      DebugMsg("Supported # of cursors: %d\n", nCursors);

    ret = pfWTInfo(WTI_INTERFACE, IFC_NCONTEXTS, &val);
    DebugMsg("Supported # of contexts: %d\n", val);

    int tabnum=0;
    while (0 != (ret = pfWTInfo(WTI_DEVICES+tabnum, DVC_NAME, strbuf)) ) {

      DebugMsg("Tablet %d name: \"%s\"\n", tabnum, LPCSTR(strbuf));

      UINT hwFlags=0;
      ret = pfWTInfo(WTI_DEVICES+tabnum, DVC_HARDWARE, &hwFlags);
      if (hwFlags & HWC_INTEGRATED) 
        DebugMsg("   Has integrated screen.\n");
      if (hwFlags & HWC_TOUCH)
        DebugMsg("   Is touch-screen variety.\n");
      if (hwFlags & HWC_HARDPROX)
        DebugMsg("   Reports enter/leave events.\n");
      if (hwFlags & HWC_PHYSID_CURSORS)
        DebugMsg("   Uses unique cursor IDs.\n");
      tabnum++;
    }
    tabnum--;
    if (tabnum < 0) {
      DebugMsg("No tablet devices found\n");
    }
  }
#else
  if (getNumDevices()==0) {
    DebugMsg("No tablet devices found\n");
  }
  else if (getNumActiveDevices()==0) {
    DebugMsg("No active tablet devices found\n");
  }
#endif

  /////////////////////////////////
  // See what kind of cursors there are
  UINT ret;
  int i;
  for(i=0; 0 != (ret=pfWTInfo( WTI_CURSORS + i, CSR_NAME, strbuf )); i++) {
    //pfWTInfo( WTI_CURSORS + i, CSR_NAME, strbuf );
    VerboseDebugMsg("Cursor %d: \"%s\"\n", i, strbuf);

    m_data->m_cursors.push_back(_bbTabletCursor());
    m_data->m_cursors[i].name = strbuf;

    {
      // browse WinTab's many info items to discover pressure handling.
      int wActiveCsr = i;
      BYTE logBtns[32];
      UINT size;

      BOOL bActive;
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_ACTIVE, &bActive);
      if (ret)
        VerboseDebugMsg("  %s.\n", bActive?"Active":"Not active");

      WTPKT pktData;
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_PKTDATA, &pktData);
#if VERBOSE_DEBUG
      if (bActive && ret) {
        DebugMsg("  Supports the following: ");
        if (pktData & PK_X) { DebugMsg("x "); }
        if (pktData & PK_Y) { DebugMsg("y "); }
        if (pktData & PK_Z) { DebugMsg("z "); }
        if (pktData & PK_BUTTONS) { DebugMsg("btns "); }
        if (pktData & PK_CONTEXT) { DebugMsg("ctx "); }
        if (pktData & PK_STATUS) { DebugMsg("stat "); }
        if (pktData & PK_TIME) { DebugMsg("tm "); }
        if (pktData & PK_CHANGED) { DebugMsg("chng "); }
        if (pktData & PK_SERIAL_NUMBER) { DebugMsg("sn "); }
        if (pktData & PK_CURSOR) { DebugMsg("csr "); }
        if (pktData & PK_NORMAL_PRESSURE) { DebugMsg("np "); }
        if (pktData & PK_TANGENT_PRESSURE) { DebugMsg("tp "); }
        if (pktData & PK_ORIENTATION) { DebugMsg("or "); }
        if (pktData & PK_ROTATION) { DebugMsg("rot "); }
        DebugMsg("\n");
      }
#endif

      DWORD physID = 0;
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_PHYSID, &physID);
      if (ret)
        VerboseDebugMsg("  physical id: %d\n", physID);

      BYTE nButtons = 0;
      pfWTInfo(WTI_CURSORS+wActiveCsr, CSR_BUTTONS, &nButtons);
      m_data->m_cursors[i].nButtons = nButtons;
      VerboseDebugMsg("  has %d buttons\n", nButtons);


      ret = pfWTInfo(WTI_CURSORS+wActiveCsr, CSR_BTNNAMES, strbuf);
      if (ret && nButtons>0) {
        VerboseDebugMsg("  Button names:\n");
        char *p = strbuf;
        for (int j=0; j<nButtons; j++) {
          VerboseDebugMsg("  %d - \"%s\"\n", j, p);
          m_data->m_cursors[i].buttonNames.push_back( string(p) );
          p+=strlen(p)+1;
        }
      }

      // discover the LOGICAL button generated by the pressure channel
      // get the PHYSICAL button from the cursor category and run it 
      // through that cursor's button map (usually the identity map). 
      BYTE wPrsBtn = (BYTE)-1;
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_NPBUTTON, &wPrsBtn);
      size = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_BUTTONMAP, &logBtns);
      if ((UINT)wPrsBtn < size)
        wPrsBtn = logBtns[wPrsBtn];

      // get the current context for its device variable. 
      // pfWTGet(hTab, &lc);

      // get the button marks (up & down generation thresholds) 
      // and calculate what pressure range we get when pressure-button is down. 

      /*
      AXIS np;
      UINT btnMarks[2];
      btnMarks[1] = 0; // default if info item not present. 
      ret = pfWTInfo(WTI_CURSORS + wActiveCsr, CSR_NPBTNMARKS, btnMarks);
      UINT prsYesBtnOrg = btnMarks[1];
      UINT prsYesBtnExt = (UINT)np.axMax - btnMarks[1];
      */
    }
  }
  m_data->_setupCallbackHooks();

  ret = reinitTablets();
  return ret && setPointerMode( mode, BBT_ALL_DEVICES );
}

bool bbTablet::reinitTablets()
{
  int numd = getNumDevices();
  bool ret = true;
  for (int i=0; i<numd; i++) {
    ret = ret && m_data->_reopenDeviceContext( i );
  }
  return ret;
}

bool bbTablet::setPointerMode( int mode, int devID /* = BBT_ALL_DEVICES  */ )
{
  // switch between SYSTEM_POINTER and SEPARATE_POINTER modes
  if (devID == BBT_ALL_DEVICES) {
    int numd = getNumDevices();
    bool ret = true;
    for (int i=0; i<numd; i++) {
      ret = ret && m_data->_setDevicePointerMode( mode, i );
    }
    return ret;
  }
  else if (devID == BBT_INTEGRATED_DEVICES) 
  {
    int numd = getNumDevices();
    bool ret = true;
    for (int i=0; i<numd; i++) {
      if (isDeviceIntegrated(i))
        ret = ret && m_data->_setDevicePointerMode( mode, i );
    }
    return ret;
  }
  else if (devID == BBT_NONINTEGRATED_DEVICES) 
  {
    int numd = getNumDevices();
    bool ret = true;
    for (int i=0; i<numd; i++) {
      if (!isDeviceIntegrated(i))
        ret = ret && m_data->_setDevicePointerMode( mode, i );
    }
    return ret;
  }
  return m_data->_setDevicePointerMode( mode, devID );
}

bool _bbTabletInternal::_setDevicePointerMode(int mode, int devID)
{
  assert(devID>=0 && devID<BBT_MAX_DEVICES);
  if (devID<0 || devID>=BBT_MAX_DEVICES) { return false; }
  if (!m_dev[devID].tabCtx) { return false; }

  LOGCONTEXT lc;
  bool enable = (mode == bbTablet::SYSTEM_POINTER);
  BOOL ret = pfWTGet(m_dev[devID].tabCtx, &lc);

  if (ret && enable && !(lc.lcOptions&CXO_SYSTEM)) {
    lc.lcOptions |= CXO_SYSTEM;
    ret = pfWTSet(m_dev[devID].tabCtx, &lc);
  }
  else if (ret && !enable && (lc.lcOptions&CXO_SYSTEM)) {
    lc.lcOptions &= ~CXO_SYSTEM;
    ret = pfWTSet(m_dev[devID].tabCtx, &lc);
  }
  if (!ret)
  {
    printf("Unable to change message delivery status on the fly.\n");
  }
  return ret!=0;
}

bool _bbTabletInternal::_reopenDeviceContext( int devID )
{
  assert(devID>=0 && devID<BBT_MAX_DEVICES);
  if (devID<0 || devID>=BBT_MAX_DEVICES) { return false; }

  LOGCONTEXT lcMine;

  if (m_dev[devID].tabCtx) 
  {
    // get the old context parameters
    pfWTGet(m_dev[devID].tabCtx, &lcMine);
    // close current context
    pfWTClose(m_dev[devID].tabCtx);
    m_dev[devID].tabCtx = 0;
  }
  else 
  {
    // Get default context parameters.  Opening context for the first time

    // UINT contextUsed = WTI_DEFSYSCTX;
    //UINT contextUsed = (mode==SYSTEM_POINTER) ? WTI_DEFSYSCTX : WTI_DEFCONTEXT;
    //UINT contextUsed = devID + ((mode==bbTablet::SYSTEM_POINTER) ? WTI_DSCTXS : WTI_DDCTXS);

    UINT contextUsed = devID + WTI_DSCTXS;

    // Get default context info
    UINT ret = pfWTInfo(contextUsed, 0, &lcMine);
    if (!ret) {
      // no support for this context type
      DebugMsg("No support for context type requested");
      return false;
    }
  }

  // Customize some fields the way we like them, no matter what they were set to.
  lcMine.lcPktData = FULLPACKETDATA;
  lcMine.lcPktMode = FULLPACKETMODE;
  lcMine.lcMoveMask = 0;
  lcMine.lcBtnUpMask = lcMine.lcBtnDnMask;

  // Change the packet data rate.  Sometimes works sometimes doesn't.
  // Actually let's just leave it at the default because this is something 
  // the user can change via Wacom's control panel.
  // Also I'm seeing a phenomenon where the rate in SEPARATE_POINTER mode is 
  // exactly half that in SYSTEM_POINTER mode for some reason.
  // it it doesn't necessarily correlate with actual measured rate.  Odd.
  // lcMine.lcPktRate = 100;

  // Use full, native sub-pixel res for digitization
  lcMine.lcOutExtX = lcMine.lcInExtX;
  lcMine.lcOutExtY = lcMine.lcInExtY;
  lcMine.lcOutExtZ = lcMine.lcInExtZ;

  // BUG: Wacom's driver has trouble figuring out proper system desktop extents
  // for Cintiq.  It seems to get stuck in whatever mode it was in when you
  // plugged the tablet in the first time, regardless of the current orientation
  // of the screen.  
  // This doesn't actually seem to fix it, but at least we
  // want to use the right values below for our own coordinate mapping purposes.
  // UPDATE: Futzing with the display control panel rotation settings seemed
  // to fix it (i.e. rotate the screen, then unrotated again)  Also deleting
  // the device from the Tablet control panel helps exorcise some ghosts, too.
  int SysExtX = GetSystemMetrics(SM_CXVIRTUALSCREEN);
  int SysExtY = GetSystemMetrics(SM_CYVIRTUALSCREEN);
  lcMine.lcSysExtX = SysExtX;
  lcMine.lcSysExtY = SysExtY;

  // Don't do packet messages unless someone is asking for callbacks
  if (_hasObservers()) {
    lcMine.lcOptions |= CXO_MESSAGES; 
  } else {
    lcMine.lcOptions &= ~CXO_MESSAGES; 
  }
  // but csrmessages are ok.
  lcMine.lcOptions |= CXO_CSRMESSAGES;

  // Set the device to use
  //  -1 seems to gather input from all devices (and is the Wacom default)
  //  or use offset id's 0,1,2... like in the WTI_DEVICES enumeration
  lcMine.lcDevice = devID;

  m_dev[devID].tabCtx = pfWTOpen( m_hwnd, &lcMine, TRUE );
  if (!m_dev[devID].tabCtx) {
    // no support for these context options
    DebugMsg("No support for context type requested");
    return false;
  }

  // Look at what we actually got (this doesn't totally work. I know it doesn't
  // because I set the report rate really high before WTOpen but nothing happened.
  // Supposedly it should be setting it to what it really is.
  if (!pfWTGet( m_dev[devID].tabCtx, &lcMine))
  {
    // no support for these context options
    DebugMsg("Couldn't get new context info. Nonfatal?");
  }

  // Turns out this value is often bogus, so there's not much point in 
  // reporting it to the user.
  //DebugMsg("Device %d Report rate: %d Hz\n", devID, lcMine.lcPktRate);

  // look up scaling for doing event coordinate mappings later
  {
    AXIS axes[3];

    pfWTInfo(WTI_DEVICES+devID, DVC_NPRESSURE, &axes[2]);
    m_dev[devID].scalePress = 1.0f / axes[2].axMax;
    pfWTInfo(WTI_DEVICES+devID, DVC_TPRESSURE, &axes[2]);
    m_dev[devID].scaleTPress = 1.0f / axes[2].axMax;

    LONG aspectx = lcMine.lcOutExtX;
    LONG aspecty = lcMine.lcOutExtY;
    m_dev[devID].scaleY = 1.0f/aspecty;
    m_dev[devID].scaleX = 1.0f/aspectx;
    m_dev[devID].AspectRatio = float(aspecty)*m_dev[devID].scaleX;

    //determineDesktopScaling();
    m_dev[devID].sysScaleX = m_dev[devID].scaleX * SysExtX;
    m_dev[devID].sysScaleY = m_dev[devID].scaleY * SysExtY;

    pfWTInfo(WTI_DEVICES+devID, DVC_X, &axes[0]);
    pfWTInfo(WTI_DEVICES+devID, DVC_Y, &axes[1]);
    //DebugMsg("Output res capability %ldx%ld\n", 
    // axes[0].axMax-axes[0].axMin,axes[1].axMax-axes[1].axMin);
    //axes[0].axResolution,axes[1].axResolution);

    m_dev[devID].scaleAng = 0;
    if (pfWTInfo(WTI_DEVICES+devID, DVC_ORIENTATION, axes))
    {
      if (axes[1].axResolution != 0)
        // axis1 seems to have valid data even if other axes don't 
        m_dev[devID].scaleAng = 2.0f *M_PIf / fix2float(axes[1].axResolution);
    }
  }
  return true;
}

int _bbTabletInternal::_getDevicePointerMode(int devID)
{
  LOGCONTEXT lctx;
  pfWTGet(m_dev[devID].tabCtx, &lctx);
  return (lctx.lcOptions & CXO_SYSTEM) ? 
    bbTablet::SYSTEM_POINTER : bbTablet::SEPARATE_POINTER ;
}

int bbTablet::pointerMode( int devID /* =0  */ ) const
{
  return m_data->_getDevicePointerMode( devID );
}


bool bbTablet::waitForNextEvent(bbTabletEvent &tev, unsigned int msec)
{
  // TODO! THIS FUNCTION DOESN'T WORK ANYMORE!!  NEED TO QUEUE ON N CONTEXTS
  if (!pfWTQueuePacketsEx) return false; // not inited?
  UINT FAR oldSN, newSN;
  struct timeb stamp, newstamp;
  ftime(&stamp);
  unsigned int elapsed = 0;
  int devID = 0; 
  while (! pfWTQueuePacketsEx(m_data->m_dev[devID].tabCtx, &oldSN, &newSN) )
  {
    ftime(&newstamp);
    elapsed = 1000 * (newstamp.time - stamp.time);
    elapsed += newstamp.millitm - stamp.millitm;
    if (elapsed > msec) return 0;
  }
  return getNextEvent(tev);
}

// getNextEvent
// @return true if there was an event, false otherwise.
bool bbTablet::getNextEvent(bbTabletEvent &tev, int devID)
{
  if (!pfWTPacketsGet) return false; // maybe not initialized yet
  assert(tev.rawDataSz >= sizeof(FULLPACKET));
  FULLPACKET &pkt = *(FULLPACKET*)tev.rawData;
  
  assert(devID >= 0 || devID == BBT_ALL_DEVICES);
  if (!(devID >= 0 || devID == BBT_ALL_DEVICES)) 
    return false;

  int ndevs = getNumDevices();
  int nqdev = 1;
  int dlo = m_data->m_nextQueryDev;
  if (devID == BBT_ALL_DEVICES) {
    nqdev = ndevs;
  } else {
    dlo = devID;
  }
  int dhi = dlo + nqdev;
  
  int got = 0;
  UINT FAR oldSN, newSN;
  int id;
  for (id=dlo; id!=dhi; id++)
  {
    devID = id%ndevs;
    pfWTOverlap(m_data->m_dev[devID].tabCtx,true); // make sure our context is on top

    if (m_data->m_throwAwayOldPackets) {
      // just throw away packets that build up. (not such a hot idea...,
      // especially if you want to use multiple pens.)
      if (pfWTQueuePacketsEx(m_data->m_dev[devID].tabCtx, &oldSN, &newSN))
      {
        got = pfWTPacket(m_data->m_dev[devID].tabCtx, newSN, &pkt);
        m_data->m_lostPackets += newSN - oldSN;
        if (got) {
          ++id;
          break;
        }
      }
    }
    else {
      got = pfWTPacketsGet(m_data->m_dev[devID].tabCtx, 1, &pkt);
      if (got) {
        ++id;
        break;
      }
    }
  }
  m_data->m_nextQueryDev=(id%ndevs);

  if (!got) return false;

  // in the test program, this doesn't trigger very often.
  //if (got>1) DebugMsg("-----!!!!  Tossed %d messages ----------\n", got-1);
  if (pkt.pkStatus&TPS_QUEUE_ERR) {
      DebugMsg("bbtablet::getNextEvent(): queue error (non-fatal)\n");
  }

  if (pkt.pkChanged)
  {
#if 0
    DebugMsg("PKChanges: ");
    if (pkt.pkChanged & PK_X)                { DebugMsg("x|"); }
    if (pkt.pkChanged & PK_Y)                { DebugMsg("y|"); }
    if (pkt.pkChanged & PK_Z)                { DebugMsg("z|"); }
    if (pkt.pkChanged & PK_BUTTONS)          { DebugMsg("b|"); }
    if (pkt.pkChanged & PK_NORMAL_PRESSURE)  { DebugMsg("np|"); }
    if (pkt.pkChanged & PK_TANGENT_PRESSURE) { DebugMsg("tp|"); }
    if (pkt.pkChanged & PK_ORIENTATION)      { DebugMsg("or|"); }
    if (pkt.pkChanged & PK_ROTATION)         { DebugMsg("rot|"); }
    if (pkt.pkChanged & PK_CONTEXT)          { DebugMsg("c|"); }
    if (pkt.pkChanged & PK_STATUS)           { DebugMsg("s|"); }
    if (pkt.pkChanged & PK_TIME)             { DebugMsg("t|"); }
    if (pkt.pkChanged & PK_SERIAL_NUMBER)    { DebugMsg("sn|"); }
    if (pkt.pkChanged & PK_CHANGED)          { DebugMsg("chng|"); }
    if (pkt.pkChanged & PK_CURSOR)           { DebugMsg("csr|"); }
    DebugMsg("\n");
#endif

    tev.deviceID = devID;

    // Scale data to [0,1]
    tev.x = pkt.pkX * m_data->m_dev[devID].scaleX;
    tev.y = pkt.pkY * m_data->m_dev[devID].scaleY;
    tev.z = pkt.pkZ * m_data->m_dev[devID].scaleZ;
    tev.pressure = pkt.pkNormalPressure * m_data->m_dev[devID].scalePress;
    tev.tpressure = pkt.pkTangentPressure * m_data->m_dev[devID].scaleTPress;
    tev.id = 0; // in case feature isn't supported
    pfWTInfo(WTI_CURSORS + pkt.pkCursor , CSR_PHYSID, &tev.id );
    tev.type = pkt.pkCursor;
    tev.buttons = pkt.pkButtons;

    // Now orientation is done on-demand by calling getOrientation,
    // because it's a lot of wasted computation if you don't actually want it.
    //if (pkt.pkChanged & PK_ORIENTATION)
    tev._flags &= ~BBTEV_ORIENT_OK;
    //  convertOrientation2AngleAxis(pkt.pkOrientation, m_data->scaleAng,
    //                               &tev.angle, tev.axis);
    return true;
  }

  return false;

}

bool bbTablet::isValid( int devID /* = BBT_ALL_DEVICES  */) const
{
  if (!m_data) return false;
  if (devID >= 0) return (m_data->m_dev[devID].tabCtx!=0);
  assert(devID == BBT_ALL_DEVICES);
  bool valid=true;
  int n = getNumDevices();
  for (int i=0; i<n; i++) {
    valid = valid && (m_data->m_dev[i].tabCtx!=0);
  }
  return valid;
}
void bbTablet::setDropOldEvents(bool onOff)
{
  if (m_data) m_data->m_throwAwayOldPackets = onOff;
}
bool bbTablet::getDropOldEvents() const
{
  return m_data ? m_data->m_throwAwayOldPackets : false;
}
long bbTablet::getNumDroppedEvents() const
{
  return m_data ? m_data->m_lostPackets : 0;
}
void bbTablet::resetNumDroppedEvents()
{
  if (m_data) m_data->m_lostPackets = 0;
}

const char *bbTablet::getCursorName(int csrtype)
{
  if (csrtype < 0 || csrtype >= (int)m_data->m_cursors.size()) {
    return 0;
  }
  return m_data->m_cursors[csrtype].name.c_str();
}
int bbTablet::getCursorNumButtons(int csrtype)
{
  if (csrtype < 0 || csrtype >= (int)m_data->m_cursors.size()) {
    return 0;
  }
  return m_data->m_cursors[csrtype].nButtons;
}
const char *bbTablet::getCursorButtonName(int csrtype, int buttonnum)
{
  if (csrtype < 0 || csrtype >= (int)m_data->m_cursors.size() || buttonnum < 0) {
    return 0;
  }
  if (buttonnum >= (int)m_data->m_cursors[csrtype].buttonNames.size()) {
    return 0;
  }
  return m_data->m_cursors[csrtype].buttonNames[buttonnum].c_str();

}

int bbTablet::getNumActiveDevices() const
{
  UINT ntabs;
  pfWTInfo(WTI_INTERFACE, IFC_NDEVICES, &ntabs);
  return ntabs;
}

int bbTablet::getNumDevices() const
{
  int tabnum=0;
  UINT hwFlags=0;
  UINT ret;
  while (0 != (ret = pfWTInfo(WTI_DEVICES+tabnum, DVC_HARDWARE, &hwFlags))) 
  {
    tabnum++;
  }
  return tabnum;
}

bool bbTablet::isDeviceIntegrated(int devID) const
{
  UINT hwFlags=0;
  pfWTInfo(WTI_DEVICES+devID, DVC_HARDWARE, &hwFlags);
  return (0!=(hwFlags & HWC_INTEGRATED));
}

/**
 * addObserver
 *    Adds a tablet event observer.  With this, users can get tablet events
 *    via callbacks rather than polling.  The observer's OnTabletEvent 
 *    method will be called whenever new events are ready.
 */
bool bbTablet::addObserver( bbTabletObserver& obs )
{
  if (m_data->m_observers.empty()) {
    m_data->_setMessagesEnabled(true);
  }
  return m_data->_addObserver(obs);
}

/**
 * removeObserver
 *    Removes a previously added tablet event observer.  
 *    Returns false if the observer couldn't be removed (generally because
 *    it was never added.)
 */
bool bbTablet::removeObserver( bbTabletObserver& obs )
{
  if (m_data->m_observers.empty())
    return false;

  bool ret = m_data->_removeObserver(obs);
  if (m_data->m_observers.empty())
    m_data->_setMessagesEnabled(false);
  return ret;
}


//============================================================================
// BBTABLET DEVICE INTERNALS 
//============================================================================

_bbTabletInternal::_bbTabletInternal()
  : m_hwnd(0), m_nextQueryDev(0)
{ 
  initWintabDLL();
}

static  void convertOrientation2AngleAxis(
    const ORIENTATION &orient,
    float scale2rads,
    float *angle,
    float axis[3])
{
    bool flip = orient.orAltitude < 0; 

    float  az = orient.orAzimuth  * scale2rads;
    float alt = orient.orAltitude * scale2rads;
    float  tw = orient.orTwist    * scale2rads;

    //   this is sort of like an euler zyz angles representation, 
    //   but a bit different since altitude depends on whether the pen
    //   is upside down or not, but azimuth doesn't care.
    //   Also we want z up to be (0,0,0) in angles, not horizontal.

    //  Remap altitude: 0 is straight up, pi is straight upside down
    alt = M_PIf*0.5f - alt;
    //  Remap azimuth: 0 is +x axis, and increases CCW, decreases CW
    if (az > M_PI*1.5f) {
      az = M_PIf*2.5f - az;
    }
    else {
      az = M_PIf*0.5f - az;
    }


    float sa = (float)sin(alt);
    float ca = (float)cos(alt);
    float st = (float)sin(tw);
    float ct = (float)cos(tw);
    float sz = (float)sin(az);
    float cz = (float)cos(az);
    if (!flip) {
      // In Matlab terms, we have the following:
      // Raz_z = [ cz -sz  0 ; sz   cz    0 ;  0   0   1 ];
      // Ral_y = [ ca  0  sa ;  0    1    0 ; -sa  0  ca ];
      // Rtw_z = [ ct -st  0 ; st   ct    0 ;  0   0   1 ];
      // R = Raz_z * Ral_y * Raz_z.' * Rtw_z;
      // That's where the trace, and rXYmrYX terms below come from

      float trace = ct*(ca+1.0f)+ca;

      float cosAngle = 0.5f*(trace - 1.0f);
      *angle = float( acos(cosAngle) );
      if (*angle == 0) {
        axis[0] = 0;
        axis[1] = 0;
        axis[2] = 1.0f;
      }
      else {
        float r32mr23 = sa*( st*cz - ct*sz - sz);
        float r21mr12 = (ca+1)*st;
        float r13mr31 = sa*( cz + ct*cz + st*sz);

        float coef = 0.5f / (float)sin(*angle);
        axis[0] = coef * (r32mr23);
        axis[1] = coef * (r13mr31);
        axis[2] = coef * (r21mr12);
      }
    }
    else { // FLIP IT!
      // Flipped case is handled differently so that there won't be a singularity
      // when the pen is exactly upside down.
      // In Matlab terms, we have the following:
      // Raz_z = [ cz -sz  0 ; sz   cz    0 ;  0   0   1 ];
      // Ral_y = [ ca  0  sa ;  0    1    0 ; -sa  0  ca ];
      // Rtw_z = [ ct -st  0 ; st   ct    0 ;  0   0   1 ];
      // roty180 = [ -1 0 0; 0 1 0; 0 0 -1 ];

      // R = roty180 * Raz_z * Ral_y * Raz_z.' * Rtw_z;
      // We also modify some of the angles:
      //   * use -az instead of az.
      //   * use 180-alt instead of alt.
      // That's where the trace, and rXYmrYX terms below come from

      sz = -sz; // az  => -az
      ca = -ca; // alt => 180 - alt

      float szcz2 = 2*sz*cz;
      float cz2msz2 = cz*cz - sz*sz;
      float va = (1.0f-ca);
      float trace = va * ( szcz2*st + cz2msz2*ct ) - ca;

      float cosAngle = 0.5f*(trace - 1.0f);
      *angle = float( acos(cosAngle) );
      if (*angle == 0) {
        axis[0] = 0;
        axis[1] = 0;
        axis[2] = 1.0f;
      }
      else {
        float r32mr23 = sa * (sz*ct - st*cz - sz);
        float r21mr12 =  va*( -szcz2*ct + cz2msz2*st ) ;
        float r13mr31 = -sa*(cz + ct*cz + sz*st);

        float coef = 0.5f / float( sin(*angle) );
        axis[0] = coef * (r32mr23);
        axis[1] = coef * (r13mr31);
        axis[2] = coef * (r21mr12);
      }
    }
}

void _bbTabletInternal::_convertOrientation2AngleAxis(
    const ORIENTATION &orient, float scaleAng, float *angle, float axis[3])
{
  ::convertOrientation2AngleAxis(orient,scaleAng,angle,axis);
}

UINT  _bbTabletInternal::ms_msgBase = 0;
HHOOK _bbTabletInternal::ms_msgFilterHook = 0;

void _bbTabletInternal::msgProcProcessWTPacket(HCTX hCtx)
{
  // Find the context that matches
  for (int devID=0; devID<BBT_MAX_DEVICES; devID++) {
    if (m_dev[devID].tabCtx==hCtx) 
    {
      while (bbTablet::getInstance().getNextEvent(m_evt, devID))
      {
        _callObservers(m_evt);
      }
      break;
    }
  }
}

void _bbTabletInternal::msgProcProcess(const MSG& msg)
{
  switch(LOWORD(msg.message))
  {
  case WT_PACKET: // The main reason we come here.

    //DebugMsg("tanoshii GetMessage! WT_PACKET\n");
    msgProcProcessWTPacket((HCTX)msg.lParam);
    break;

  case WT_PROXIMITY:		
    //DebugMsg("tanoshii GetMessage! WT_PROXIMITY\n");
    // A pen entered or left the vicinity of a tablet.
    // Usually before any actual contact occurs.
    // This might be worth making a callback.
    break;
  case WT_CSRCHANGE:
    //DebugMsg("tanoshii GetMessage! WT_CSRCHANGE\n");
    // Happens when the pen near the tablet changes.
    // This might be worth making a callback.
    // It might make the users life easier in dealing with multiple pen input
    break;
  case WT_INFOCHANGE:	
    // DebugMsg("tanoshii GetMessage! WT_INFOCHANGE\n");
    // This is like when a new tablet is plugged in.  Run
    // through initialization again when this happens.
    // (Is this sufficient??  Callback might be useful here too)
    bbTablet::getInstance().reinitTablets();
    break;
#if 0
  case WT_CTXOPEN:
    //DebugMsg("tanoshii GetMessage! WT_CTXOPEN\n");
    // A context was opened.
    break;
  case WT_CTXCLOSE:
    //DebugMsg("tanoshii GetMessage! WT_CTXCLOSE\n");
    // A context was closed.
    break;
  case WT_CTXUPDATE:		
    //DebugMsg("tanoshii GetMessage! WT_CTXUPDATE\n");
    // Not documented, and I've never seen it called.  Maybe if the user
    // futzes with the control panel while your app is open?
    break;
  case WT_CTXOVERLAP:		
    //DebugMsg("tanoshii GetMessage! WT_CTXOVERLAP\n");
    // something changed in the stacking order of tablet contexts.
    // Not sure how useful this is, really.
    break;
#endif
  }
}

LRESULT CALLBACK _bbTabletInternal::msgProcFilter(
  int nCode, WPARAM wParam, LPARAM lParam)
{
  if (nCode<0)
    return CallNextHookEx(ms_msgFilterHook,nCode,wParam,lParam);

  _bbTabletInternal *obj = bbTablet::getInstance().m_data;
  if (nCode == HC_ACTION)
  {
    MSG* pMsg = (MSG*)lParam;

    //if (pMsg->hwnd == obj->hwnd)
    {
      //if (LOWORD(pMsg->message) >= s_msgBase && LOWORD(pMsg->message)<=s_msgBase+WT_MAXOFFSET )
      if (LOWORD(pMsg->message) >= WT_PACKET && LOWORD(pMsg->message) <= WT_MAX)
      {
        obj->msgProcProcess(*pMsg);
      }
    }
  }
  return CallNextHookEx(ms_msgFilterHook, nCode, wParam, lParam);
}

void _bbTabletInternal::_setupCallbackHooks()
{
  pfWTInfo(WTI_DEFCONTEXT, CTX_MSGBASE, &ms_msgBase);
  ms_msgFilterHook = SetWindowsHookEx(WH_GETMESSAGE,msgProcFilter,NULL,GetCurrentThreadId());
  if (ms_msgFilterHook==NULL) {
    printf("COuldn't hook");
    //FormatMessage(GetLastError()
  }
}
void _bbTabletInternal::_removeCallbackHooks()
{
  UnhookWindowsHookEx(ms_msgFilterHook);
  ms_msgFilterHook = 0;
}

void _bbTabletInternal::_setMessagesEnabled(bool enable)
{
  LOGCONTEXT lc;
  int ntabs = bbTablet::getInstance().getNumDevices();
  for (int i=0; i<ntabs; i++)
  {
    if (!m_dev[i].tabCtx) 
      continue;
    BOOL ret = pfWTGet(m_dev[i].tabCtx, &lc);
    if (ret && enable && !(lc.lcOptions&CXO_MESSAGES)) {
      lc.lcOptions |= CXO_MESSAGES;
      ret = pfWTSet(m_dev[i].tabCtx, &lc);
      if (ret) {
        // flush the queue of anything that was left over also.  just for fun.
        pfWTPacketsGet(m_dev[i].tabCtx, 1, NULL);
      }
    }
    else if (ret && !enable && (lc.lcOptions&CXO_MESSAGES)) {
      lc.lcOptions &= ~CXO_MESSAGES;
      ret = pfWTSet(m_dev[i].tabCtx, &lc);
    }
    if (!ret)
    {
      printf("Unable to change message delivery status on the fly.\n");
    }
  }
}
//============================================================================
// BBTABLET EVENT
//============================================================================

bbTabletEvent::bbTabletEvent() : _flags(0) {
  rawDataSz = sizeof(FULLPACKET);
  rawData = malloc(rawDataSz);
}
bbTabletEvent::~bbTabletEvent() {
  if (rawData) free(rawData);
  rawData = 0;
  rawDataSz = 0;
}
bbTabletEvent::bbTabletEvent(const bbTabletEvent &o)
{
  *this = o;
}

bbTabletEvent& bbTabletEvent::operator=(const bbTabletEvent &o)
{
  if (&o==this) 
    return *this;
  void *newData = rawData;
  if (rawDataSz != o.rawDataSz) {
    if (!rawDataSz) { 
      newData = malloc(o.rawDataSz);
    }
    else if (!o.rawDataSz) {
      free(rawData);
      newData = 0;
    }
    else {
      free(rawData);
      newData = malloc(o.rawDataSz);
    }
  }
  memcpy(this, &o, sizeof(bbTabletEvent));
  if (rawDataSz) memcpy(newData, o.rawData, o.rawDataSz);
  rawData = newData;
  return *this;
}

void bbTabletEvent::getOrientation(float &angle, float axis[3]) const
{
  if (!(_flags & BBTEV_ORIENT_OK)) 
  {
    FULLPACKET &pkt = *(FULLPACKET*)rawData;
    //if (pkt.pkChanged & PK_ORIENTATION)
    _bbTabletInternal *data = bbTablet::getInstance().m_data;
    data->_convertOrientation2AngleAxis(
      pkt.pkOrientation, data->m_dev[deviceID].scaleAng, &_angle, _axis);
    _flags |= BBTEV_ORIENT_OK;
  }
  angle = _angle;
  memcpy(axis,_axis,3*sizeof(_axis[0]));
}
int bbTabletEvent::getRawX() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkX;
}
int bbTabletEvent::getRawY() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkY;
}
int bbTabletEvent::getRawZ() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkZ;
}
float bbTabletEvent::getSysXf() const
{
  // Use full virtual desktop coords
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkX * bbTablet::getInstance().m_data->m_dev[deviceID].sysScaleX;
}
float bbTabletEvent::getSysYf() const
{
  // Use virtual desktop coords
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  float y = pkt.pkY * bbTablet::getInstance().m_data->m_dev[deviceID].sysScaleY;
  // Needs to be inverted to match windows.  Tablets seem to have lower left origin
  // instead of Windows' upper left.
  return GetSystemMetrics(SM_CYVIRTUALSCREEN)-1 - y;
}
float bbTabletEvent::getWinXf(void *win_handle) const
{
  POINT pt = { 0, 0}; 
  // ASSUMING THERE'S UNIFORM SCALING!
  MapWindowPoints(NULL, (HWND)win_handle, &pt, 1);
  return getSysXf() + pt.x;
}
float bbTabletEvent::getWinYf(void *win_handle) const
{
  POINT pt = { 0, 0}; 
  // ASSUMING THERE'S UNIFORM SCALING!
  MapWindowPoints(NULL, (HWND)win_handle, &pt, 1);
  return getSysYf() + pt.y;
}
float bbTabletEvent::getWinXf() const
{
  return getWinXf(bbTablet::getInstance().m_data->m_hwnd);
}
float bbTabletEvent::getWinYf() const
{
  return getWinYf(bbTablet::getInstance().m_data->m_hwnd);
}

int bbTabletEvent::getRawPressure() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkNormalPressure;
}
int bbTabletEvent::getRawTangentialPressure() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkTangentPressure;
}
int bbTabletEvent::getRawOrientAzimuth() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkOrientation.orAzimuth;
}
int bbTabletEvent::getRawOrientAltitude() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkOrientation.orAltitude;
}
int bbTabletEvent::getRawOrientTwist() const 
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkOrientation.orTwist;
}
int bbTabletEvent::getRawTiltX() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkTilt.tiltX;
}
int bbTabletEvent::getRawTiltY() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkTilt.tiltY;
}
int bbTabletEvent::getRawRoll() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkRotation.roRoll;
}
int bbTabletEvent::getRawPitch() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkRotation.roPitch;
}
int bbTabletEvent::getRawYaw() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkRotation.roYaw;
}
bool bbTabletEvent::inverted() const
{
  FULLPACKET &pkt = *(FULLPACKET*)rawData;
  return pkt.pkOrientation.orAltitude<0;
}

