//============================================================================
// Bare-Bones Tablet Interface for C++
//
//  This header was designed to be platform-neutral for portability, 
//  but currently Win32 is the only implemented back end.
//
//  Relies on Wintab32.dll for windows operation.
//  If you have a tablet, then you should have a copy of Wintab32.dll that
//  was installed by the tablet's driver installer.
//  
//  Also necessary (for development on Windows) will be the WINTAB SDK.
//  A link can be found at http:://www.billbaxter.com/projects/bbtablet
//  It contains the necessary headers: wintab.h, pktdef.h ...
//
//  Author:        William Baxter (baxter@cs.unc.edu)
//  Last Modified: Dec 2005
//============================================================================
// (MIT License)
// Copyright (c) 2002--2009 William Baxter
// 
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
// BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
// ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//============================================================================


#ifndef BB_TABLET_HEADER
#define BB_TABLET_HEADER

#define BBTAB_NOAUTOLINK
// autolinking causes troubles because we can't reliably detect when the
// app is linking against the DLL Runtime Libs.  The ifdef _DLL below doesn't
// do the trick.
#if defined(_WIN32) && !defined(BBTAB_NOAUTOLINK)
#ifdef _DLL
 // Decide what version of bbtablet.lib to link with and auto-link it.
 // The mnemonic is 'd' for debug, 's' for single-threaded runtime library
 // (NOTE: This convention just changed as of 2005-12!)
 #ifdef _DEBUG
 #pragma comment (lib, "bbtabletd.lib") // for /MDd apps
 #else
 #pragma comment (lib, "bbtablet.lib") // for /MD apps
 #endif
#else 
 #ifdef _DEBUG
 #pragma comment (lib, "bbtabletds.lib") // for /MTd, /MLd apps
 #else 
 #pragma comment (lib, "bbtablets.lib") // for /ML, /MT apps
 #endif
#endif
#endif


//============================================================================
/**
 * @class bbTabletEvent
 */
class bbTabletEvent
{
public:
  float x, y, z;    // Normalized [0,1] x,y,z coords
  float pressure;   // Normalized [0,1] pressure
  float tpressure;  // "Tangential pressure" like from airbrush wheel
  unsigned int id;      // Unique physical id of stylus if supported
  unsigned int buttons; // Bitmask of pressed buttons
  int type; // The type of stylus reported by WinTab
  int deviceID; // Who generated this event?

  void getOrientation(float &angle, float axis[3]) const;
  const char *getCursorName() const;
  int getCursorNumButtons() const;
  const char *getCursorButtonName(int buttonnum) const;
  
  // Routines to get the raw values reported by the driver
  int getRawX() const;
  int getRawY() const;
  int getRawZ() const;

  float getSysXf() const; // x in (virtual) desktop coords
  float getSysYf() const;
  float getWinXf() const; // x in window coords of tablet HWND
  float getWinYf() const;
  float getWinXf(void *win_handle) const; // x in coords of given window (HWND)
  float getWinYf(void *win_handle) const; 
  int getSysX() const { return int(getSysXf()+0.5f); } // These truncate to ints
  int getSysY() const { return int(getSysYf()+0.5f); } //  but the actual tablet
  int getWinX() const { return int(getWinXf()+0.5f); } //  precision is often sub-
  int getWinY() const { return int(getWinYf()+0.5f); } //  pixel.
  int getWinX(void *win_handle) const { return int(getWinXf(win_handle)+0.5f); }  
  int getWinY(void *win_handle) const { return int(getWinYf(win_handle)+0.5f); } 

  int getRawPressure() const;
  int getRawTangentialPressure() const;
  int getRawOrientAzimuth() const;
  int getRawOrientAltitude() const;
  int getRawOrientTwist() const;
  int getRawTiltX() const;
  int getRawTiltY() const;
  int getRawRoll() const; // does anything support this?
  int getRawPitch() const; // does anything support this?
  int getRawYaw() const; // does anything support this?

  bool inverted() const; // is the stylus inverted (i.e. "eraser mode")

  void *rawData;
  size_t rawDataSz;

public:
  bbTabletEvent();
  ~bbTabletEvent();
  bbTabletEvent(const bbTabletEvent &o);
  bbTabletEvent& operator=(const bbTabletEvent &o);
private:
  mutable unsigned int _flags;
  mutable float _angle;
  mutable float _axis[3];
  friend class bbTablet;
};

enum { 
  BBT_ALL_DEVICES=-1,
  BBT_INTEGRATED_DEVICES=-2,
  BBT_NONINTEGRATED_DEVICES=-3
};

class _bbTabletInternal;

//============================================================================
/* @class bbTabletObserver
 *
 *  This class defines the callback interface for using bbTablet via callbacks
 *  rather than polling.  
 */
class bbTabletObserver
{
public:
  virtual int OnTabletEvent(const bbTabletEvent &evt) = 0;
};

//============================================================================
/**
 * @class bbTablet
 *
 *   This class is a singleton. If multiple simultaneous tablets are supported
 *   by the drivers, this will be able get all their events.
 */
class bbTablet
{
private:
  // bbTablet is a singleton, so it can't be user-instantiated or copied
  // Use getInstance() instead.
  bbTablet(); 
  bbTablet(const bbTablet&);
  ~bbTablet();

public:
  enum { SYSTEM_POINTER, SEPARATE_POINTER };

  static bbTablet &getInstance();

  /** 
   * initSystem
   * @param system_window  should be an HWND on Windows, e.g.
   * @param mode 
   *      SYSTEM_POINTER if this tablet is to drive  the system cursor
   *      SEPARATE_POINTER if this tablet doesn't affect the system cursor
   * @return true if successfully initialized
   */
  bool initSystem( void *system_window, int mode = SYSTEM_POINTER );
  
  /** 
   * initTablet
   * @deprecated Call initSystem instead, which is all this method does now.
   */
  bool initTablet( void *system_window, int mode = SYSTEM_POINTER ) 
  { return initSystem(system_window, mode); }


  /** 
   * reinitTablets
   * Close and reopen all the tablet contexts in use.  May help if something
   * changes in the tablet environment.  Basically, if things go wonky, try 
   * calling this.  But generally there's no need.
   */
  bool reinitTablets();

  /**
   * setPointerMode
   * @param mode
   *      SYSTEM_POINTER if this tablet is to drive  the system cursor
   *      SEPARATE_POINTER if this tablet doesn't affect the system cursor
   * @param devID
   *      The device ID.  If BBT_ALL_DEVICES set mode for all devices
   *                      If BBT_INTEGRATED_DEVICES set mode for all devices with screens
   *                      If BBT_NONINTEGRATED_DEVICES set mode for all devices without screens
   * @return true if pointer mode successfully set
   */
  bool setPointerMode( int mode, int devID = BBT_ALL_DEVICES );
  /**
   * pointerMode
   * Returns the current pointer mode, either SYSTEM_POINTER or SEPARATE_POINTER
   */
  int pointerMode( int devID=0 ) const;

  /**
   * addObserver
   *    Adds a tablet event observer.  With this, users can get tablet events
   *    via callbacks rather than polling.  The observer's OnTabletEvent 
   *    method will be called whenever new events are ready.  Returns false
   *    if the observer couldn't be added (generally because it was already 
   *    added previously)
   */
  bool addObserver( bbTabletObserver& obs );

  /**
   * removeObserver
   *    Removes a previously added tablet event observer.  
   *    Returns false if the observer couldn't be removed (generally because
   *    it was never added.)
   */
  bool removeObserver( bbTabletObserver& obs );

  /**
   * isValid
   *    Returns true if the tablet has been successfully initialized
   */
  bool isValid( int devID = BBT_ALL_DEVICES ) const;

  /**
   * setDropOldEvents
   *    Sets whether getNextEvent will drop events if called when there is
   *    more than one event waiting in the queue.  Off by default.  Turning this on is not 
   *    recommended if you have multiple tablets or plan to use multiple 
   *    styluses simultaneously.
   */
  void setDropOldEvents(bool onOff);
  bool getDropOldEvents() const;

  /**
   * getNumDroppedEvents
   *     If setDropOldEvents() is on, then this will return the number of
   *     events that were dropped.
   */
  long getNumDroppedEvents() const;
  void resetNumDroppedEvents();
  
  bool getNextEvent(bbTabletEvent &evt, int devID = BBT_ALL_DEVICES);
  bool waitForNextEvent(bbTabletEvent &evt, unsigned int msec = 0);

  const char *getCursorName(int csrtype);
  int getCursorNumButtons(int csrtype);
  const char *getCursorButtonName(int csrtype, int buttonnum);

  int getNumActiveDevices() const;

  int getNumDevices() const;
  bool isDeviceIntegrated(int devID) const;


private:
  _bbTabletInternal *m_data;
  friend class bbTabletEvent;
  friend class _bbTabletInternal;
};

// For backwards compatibility with previous versions of bbTablet
typedef bbTablet bbTabletDevice;

#endif /* BB_TABLET_HEADER */
