/*
 * testgl_simple.cpp
 *
 * A relatively simple OpenGL/GLUT app demonstrating some of the main features 
 * of bbTablet.  
 *
 * This doesn't really take any advantage of OpenGL, just GL+GLUT is pretty
 * much the simplest, lightest weight windowing toolkit I know.  Do a web
 * search for GLUT if you need the headers and libs for it.
 *
 * Demonstrated here are 
 * 1) on-the-fly switching between mouse-capture mode and separate input mode.
 * 2) working with bbTablet via both available techniques, POLLING and 
 *    CALLBACKS (and also with on-the-fly switching between them as well.)
 * 3) Basic info contained in a tablet event and how to access it.
 *
 * What's not demonstrated is how to manage information coming from 
 * potentially multiple tablets and multiple pens.  For that, see testgl.cpp.
 *
 * Note that the callbacks do not seem to work for console-only apps.  For 
 * those, you'll have to use polling.  I tried it in test1 first but it never
 * worked.
 *
 *  LICENSE:
 *  I place this sample code in the PUBLIC DOMAIN
 */

#include <GL/glut.h>
#include "bbtablet.h"
#include "gltext.hpp"
#include "stopwatch.hpp"
#include <stdio.h>

static bbTabletEvent g_evt;
static int g_evtFromCallback = -1;
static char g_statusMsg[1024]="hi";
static bool g_callbackInstalled = false;
static int g_mouseX=0,g_mouseY=0;
static int g_cursors[]={GLUT_CURSOR_RIGHT_ARROW, GLUT_CURSOR_CROSSHAIR, GLUT_CURSOR_NONE};
static int g_ncursors=sizeof(g_cursors)/sizeof(g_cursors[0]);
static int g_ccursor = 0;
static Stopwatch g_timer;
static unsigned int g_numEvents;
static float g_lastAvgUpdateRate = 0;

void pollTabletEvents()
{
  if (g_callbackInstalled)
    return;
  bbTabletDevice &td = bbTabletDevice::getInstance();
  
  while (td.getNextEvent(g_evt)) {
    g_evtFromCallback = false;
    g_numEvents++;
  }
}
class MyTabletObserver : public bbTabletObserver
{
public:
  int OnTabletEvent(const bbTabletEvent &evt)
  {
    g_evt = evt;
    g_evtFromCallback = true;
    g_numEvents++;
    return 0;
  }
};
MyTabletObserver g_callback;

#ifdef _WIN32
#include <windows.h>
HWND getCurrentGLWinHWND() 
{
  // get current DC from wgl
  HDC gldc = wglGetCurrentDC();
  // obtain the HWND from that
  HWND win = WindowFromDC(gldc);
  return win;
}
#endif

void drawHelp()
{
  sprintf(g_statusMsg,
    "\
Keys:\n\
's' - %s\n\
'c' - Switch to event %s mode\n\
'd' - Change the system cursor's shape\n\
",
    bbTablet::getInstance().pointerMode()==bbTablet::SYSTEM_POINTER
    ?"Release control of mouse (system pointer)"
    :"Take control of mouse (system pointer)",
    g_callbackInstalled ? "'polling'" : "'callbacks'"
    );
  glColor3f(1,1,1);
  Text(10,4*GetFontHeight(),g_statusMsg);
}

void drawWelcome()
{
  sprintf(g_statusMsg,
    "\
                         TABLET TEST\n\n\
The current stylus position is drawn as a white plus sign.\n\
The current system pointer position is shown for reference as a\n\
single green dot. Initially all tablets are in 'separate pointer'\n\
mode, i.e. styluses do not control the system pointer, so the green\n\
dot and white plus should move independently.  \n\n\
By hitting the key indicated below you can switch to controlling the\n\
system pointer with the tablet also. When doing so, the green dot\n\
should fall in the center of the plus sign if things are working\n\
properly.\
"
    );
  glColor3f(1,1,1);
  Text(10,glutGet(GLUT_WINDOW_HEIGHT)-GetFontHeight()-10,g_statusMsg);
  drawHelp();
}

void drawStats()
{
  if (g_evtFromCallback<0)
  {
    drawWelcome();
    return;
  }

  if (g_timer.GetTime()>1.0) {
    float updatePeriod = g_timer.GetTime();
    if (g_numEvents == 0) {
      g_lastAvgUpdateRate = 0;
    } else {
      g_lastAvgUpdateRate = g_numEvents/updatePeriod;
    }
    g_timer.Reset();
    g_numEvents=0;
  }
  float angle, axis[3];
  g_evt.getOrientation(angle, axis);

  sprintf(g_statusMsg,
    "\
LAST TABLET EVENT (via %s):\n\
stylus id = 0x%x\n\
cursor type = %d\n\
cursor name = %s\n\
button mask = %x\n\
position = (%0.3f, %0.3f, %0.3f)\n\
desktop position = (%0.3f, %0.3f)\n\
window  position = (%0.3f, %0.3f)\n\
pressure = %0.3f\n\
tangent pressure = %0.3f\n\
angle,axis = %0.3f, (%0.3f,%0.3f,%0.3f)\n\
Approx event rate = %0.1f Hz\n\
",
    g_evtFromCallback ? "CALLBACK" : "POLLING",
    g_evt.id,
    g_evt.type,
    g_evt.getCursorName(),
    g_evt.buttons,
    g_evt.x,
    g_evt.y,
    g_evt.z,
    g_evt.getSysXf(),
    g_evt.getSysYf(),
    g_evt.getWinXf(),
    g_evt.getWinYf(),
    g_evt.pressure,
    g_evt.tpressure,
    angle,
    axis[0],axis[1],axis[2],
    g_lastAvgUpdateRate
    );
  glColor3f(1,1,1);
  Text(10,glutGet(GLUT_WINDOW_HEIGHT)-GetFontHeight()-10,g_statusMsg);

  drawHelp();
}

void idleFunc()
{
  pollTabletEvents();
  glutPostRedisplay();
}

void drawMouseCursor()
{
  glBegin(GL_POINTS);
  glColor3f(.1,1,.1);
  glVertex2f(g_mouseX, g_mouseY);
  glEnd();
}

void drawTabletCursor()
{
  bbTablet &tablet = bbTablet::getInstance();
  float wx,wy;
  if (tablet.isDeviceIntegrated(g_evt.deviceID) ||
      tablet.pointerMode(g_evt.deviceID)==bbTablet::SYSTEM_POINTER)
  {
    // For tablets with a built-in screen or anytime in system pointer mode
    // use the window coordinates.  Use float versions because tablets are
    // often sub-pixel accurate.
    wx = g_evt.getWinXf();
    wy = g_evt.getWinYf();
  }
  else
  {
    // For non-display tablets in "SEPARATE_POINTER" mode, map the whole
    // tablet area onto the window's rectangle.  This gives denser tablet
    // sampling over the window's area.  (It is fine to use the Win 
    // coordinates, too, though.)
    wx = g_evt.x * glutGet(GLUT_WINDOW_WIDTH); 
    wy = (1.0-g_evt.y) * glutGet(GLUT_WINDOW_HEIGHT);
  }

  glColor3f(1,1,1);
  glBegin(GL_POINTS);
  glVertex2f(wx,wy);
  glEnd();

  glBegin(GL_LINES);
  glVertex2f(wx,  wy+8);
  glVertex2f(wx,  wy+3);

  glVertex2f(wx,  wy-8);
  glVertex2f(wx,  wy-3);

  glVertex2f(wx+8,wy);
  glVertex2f(wx+3,wy);
             
  glVertex2f(wx-8,wy);
  glVertex2f(wx-3,wy);

  glEnd();
}
void displayFunc()
{
  glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);

  drawStats();

  drawTabletCursor();

  drawMouseCursor();

  glutSwapBuffers();
}

void keyboardFunc(unsigned char key, int x, int y)
{
  bbTablet &tablet = bbTablet::getInstance();
  if (key=='s') {
    if (tablet.pointerMode()==bbTablet::SYSTEM_POINTER)
      tablet.setPointerMode(bbTablet::SEPARATE_POINTER);
    else
      tablet.setPointerMode(bbTablet::SYSTEM_POINTER);
  }
  else if (key=='c') {
    if (g_callbackInstalled) {
      tablet.removeObserver( g_callback );
    } else {
      tablet.addObserver( g_callback );
    }
    g_callbackInstalled = !g_callbackInstalled;
  }
  else if (key=='d') {
    ++g_ccursor %= g_ncursors;
    glutSetCursor(g_cursors[g_ccursor]);
  }
  else if (key==27) {
    exit(0);
  }
}

void specialFunc(int key, int x, int y)
{
}

void mouseFunc(int button, int state, int x, int y)
{

}

void motionFunc(int x, int y)
{
  g_mouseX = x;
  g_mouseY = y;
}

void reshapeFunc(int W, int H) 
{
  glViewport(0, 0, W, H);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,W,H,0,1,-1);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  static bool firstTime=true;
  if (firstTime) {
    void *wnd = getCurrentGLWinHWND();
    bbTabletDevice &td = bbTabletDevice::getInstance( );
    td.initSystem( wnd, bbTabletDevice::SEPARATE_POINTER );
  }

}

void placeWindow(int W, int H)
{
  int scrnW = glutGet(GLUT_SCREEN_WIDTH);
  int scrnH = glutGet(GLUT_SCREEN_HEIGHT);
  if (scrnW<W) W=scrnW;
  if (scrnH<H) H=scrnH;
  glutInitWindowPosition((scrnW-W)/2, (scrnH-H)/2);
  glutInitWindowSize(W,H);
}

int main(int argc, char* argv[])
{
  glutInit(&argc,argv);

  glutInitDisplayMode(GLUT_DOUBLE|GLUT_DEPTH);

  placeWindow(640,480);
  int winid = glutCreateWindow("bbTablet Simple GL Test");

  glutIdleFunc(idleFunc);
  glutDisplayFunc(displayFunc);
  glutKeyboardFunc(keyboardFunc);
  glutSpecialFunc(specialFunc);
  glutReshapeFunc(reshapeFunc);
  glutPassiveMotionFunc(motionFunc);
  glutMotionFunc(motionFunc);

  glClearColor(.2,.3,.3,1);


  glDisable(GL_DEPTH_TEST);

  g_timer.Start();

  glutMainLoop();

  return 0;
}
